import sys
import multiprocessing
import multiprocessing.context
import multiprocessing.pool


class SpawnNonDaemonicProcess(multiprocessing.context.SpawnProcess):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class SpawnNonDaemonicContext(multiprocessing.context.SpawnContext):
    Process = SpawnNonDaemonicProcess


if sys.platform != "win32":

    class ForkNonDaemonicProcess(multiprocessing.context.ForkProcess):
        @property
        def daemon(self):
            return False

        @daemon.setter
        def daemon(self, value):
            pass

    class ForkServerNonDaemonicProcess(multiprocessing.context.ForkServerProcess):
        @property
        def daemon(self):
            return False

        @daemon.setter
        def daemon(self, value):
            pass

    class ForkNonDaemonicContext(multiprocessing.context.ForkContext):
        Process = ForkNonDaemonicProcess

    class ForkServerNonDaemonicContext(multiprocessing.context.ForkServerContext):
        Process = ForkServerNonDaemonicProcess

    if sys.platform == "darwin":
        NONDAEMONIC_CONTEXTS = {
            None: SpawnNonDaemonicContext,
            "spawn": SpawnNonDaemonicContext,
            "fork": ForkNonDaemonicContext,
            "forkserver": ForkServerNonDaemonicContext,
        }
    else:
        NONDAEMONIC_CONTEXTS = {
            None: ForkNonDaemonicContext,
            "spawn": SpawnNonDaemonicContext,
            "fork": ForkNonDaemonicContext,
            "forkserver": ForkServerNonDaemonicContext,
        }

else:
    NONDAEMONIC_CONTEXTS = {
        None: SpawnNonDaemonicContext,
        "spawn": SpawnNonDaemonicContext,
    }


class ProcessPool(multiprocessing.pool.Pool):
    """Pool that can also manage non-daemonic processes.

    Processes from multiprocessing are daemonic by default which means:
    * the subprocess is automatically terminated after the parent process
      ends to prevent orphan processes
    * the subprocess cannot start another subprocess

    By default it uses daemonic processes which cannot have subprocesses
    and "fork" context (on Linux) or "spawn" context (on Win32).

    Non-daemonic processes
    See : https://stackoverflow.com/a/53180921
    """

    # We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
    # because the latter is only a wrapper function, not a proper class.

    def __init__(self, daemon=True, context=None, **kwargs):
        if daemon:
            context = multiprocessing.get_context(context)
        elif context in NONDAEMONIC_CONTEXTS:
            context_class = NONDAEMONIC_CONTEXTS.get(context)
            context = context_class()
        super().__init__(context=context, **kwargs)

    def join(self):
        """Wait for all workers (subprocesses) to terminate.
        Call this after `terminate` or the process might not exit.
        """
        try:
            super().join()
        except RuntimeError as e:
            if str(e) != "cannot join current thread":
                raise

    def terminate(self, wait=False):
        super().terminate()
        if wait:
            self.join()


def apply_async(
    func,
    args=tuple(),
    kwds=None,
    callback=None,
    error_callback=None,
    pool=None,
    daemon=True,
    context=None,
):
    """Launch a function in a subprocess with callbacks.

    By default it uses a daemonic process which cannot have subprocesses
    and "fork" context (on Linux) or "spawn" context (on Win32).
    """
    if kwds is None:
        kwds = dict()

    if pool:
        return pool.apply_async(
            func, args=args, kwds=kwds, callback=callback, error_callback=error_callback
        )

    pool = ProcessPool(processes=1, daemon=daemon, context=context)

    if callback is None:

        def _callback(return_value):
            pool.terminate(wait=True)

    else:

        def _callback(return_value):
            try:
                return callback(return_value)
            finally:
                pool.terminate(wait=True)

    if callback is None:

        def _error_callback(exception):
            pool.terminate(wait=True)

    else:

        def _error_callback(exception):
            try:
                return error_callback(exception)
            finally:
                pool.terminate(wait=True)

    future = pool.apply_async(
        func, args=args, kwds=kwds, callback=_callback, error_callback=_error_callback
    )
    pool.close()
    return future



import os
from typing import Optional
from .pymongo import PyMongoWorkflowDbClient


class PyBesWorkflowDbClient(PyMongoWorkflowDbClient, register_name="pybes"):
    """Client of the PyBes Mongo database for storing workflow executions."""

    def connect(self, url: Optional[str] = None):
        if not url:
            url = os.environ.get("PYPUSHFLOW_MONGOURL", None)
        super().connect(url=url, database="pybes", collection="pybes")

    def _generateInitialWorkflowInfo(self) -> dict:
        initiator = os.environ.get("PYPUSHFLOW_INITIATOR", "Unknown")
        host = os.environ.get("PYPUSHFLOW_HOST", "Unknown")
        port = os.environ.get("PYPUSHFLOW_PORT", "Unknown")
        requestId = os.environ.get("PYPUSHFLOW_OBJECTID", "Unknown")
        return {
            "initiator": initiator,
            "host": host,
            "port": port,
            "Request ID": requestId,
        }

    def generateWorkflowId(self, oid: Optional[str] = None):
        if not oid:
            oid = os.environ.get("PYPUSHFLOW_OBJECTID")
        super().generateWorkflowId(oid=oid)



import unittest
from time import sleep
from threading import Lock
from pypushflow.ThreadCountingActor import ThreadCountingActor
from pypushflow.ThreadCounter import ThreadCounter
from concurrent.futures import ThreadPoolExecutor


class Counter:
    def __init__(self):
        self.value = 0
        self._lock = Lock()

    def increment(self):
        with self._lock:
            self.value += 1


class MyThreadCountingActor(ThreadCountingActor):
    def __init__(self, thread_counter, downstream_actors=tuple()):
        super().__init__(thread_counter=thread_counter, name="MyThreadCountingActor")
        self.downstream_actors = downstream_actors

    def trigger(self, state):
        sleep(0.01)
        if not self.downstream_actors:
            state["ntasks"].increment()
        for actor in self.downstream_actors:
            actor.trigger(state)


class TestThreadCountingActor(unittest.TestCase):
    def setUp(self):
        self.thread_counter = ThreadCounter()

    def test_multiple_threads(self):
        workers1 = [MyThreadCountingActor(self.thread_counter) for _ in range(3)]
        workers2 = [
            MyThreadCountingActor(self.thread_counter, workers1) for _ in range(5)
        ]
        workers3 = [
            MyThreadCountingActor(self.thread_counter, workers2) for _ in range(10)
        ]
        state = {"ntasks": Counter()}
        with ThreadPoolExecutor(max_workers=10) as executor:
            executor.map(lambda w: w.trigger(state), workers3)
            self.thread_counter.wait_threads_finished()
            self.assertEqual(state["ntasks"].value, 150)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import pprint
from pypushflow.AbstractActor import AbstractActor


class RouterActor(AbstractActor):
    def __init__(
        self,
        parent=None,
        errorHandler=None,
        name="Router",
        itemName=None,
        listPort=None,
        **kw,
    ):
        super().__init__(parent=parent, name=name, **kw)
        self.errorHandler = errorHandler
        self.name = name
        self.itemName = itemName
        if listPort is None:
            self.listPort = []
        else:
            self.listPort = listPort
        self.dictValues = {}

    def connect(self, actor, expectedValue="other"):
        self.logger.debug(
            "connect to actor '%s' (output port %s)", actor.name, expectedValue
        )
        if expectedValue != "other" and expectedValue not in self.listPort:
            raise RuntimeError(
                f"Port {expectedValue} not defined for router actor {self.name}!"
            )
        if expectedValue in self.dictValues:
            self.dictValues[expectedValue].append(actor)
        else:
            self.dictValues[expectedValue] = [actor]

    def trigger(self, inData):
        self.logger.info("triggered with inData =\n %s", pprint.pformat(inData))
        self.setStarted()
        self.setFinished()
        listActor = None
        if self.itemName in inData:
            self.logger.debug("router item = '%s'", self.itemName)
            value = inData[self.itemName]
            self.logger.debug("router item = '%s' value = %s", self.itemName, value)
            if value in [None, "None", "null"]:
                value = "null"
            elif type(value) == bool:
                if value:
                    value = "true"
                else:
                    value = "false"
            if not isinstance(value, dict) and value in self.dictValues:
                listActor = self.dictValues[value]
        if listActor is None:
            self.logger.debug("no router destinations for inData")
            if "other" in self.dictValues:
                listActor = self.dictValues["other"]
            else:
                raise RuntimeError(f"No 'other' port for router actor '{self.name}'")
        for actor in listActor:
            actor.trigger(inData)



from typing import Callable, Optional


class WorkflowDbClient:
    """Client interface of a database for storing workflow executions."""

    _REGISTRY = dict()
    _ACTORINFO_FILTERS = list()

    def __init_subclass__(cls, register_name=None) -> None:
        super().__init_subclass__()
        if register_name:
            WorkflowDbClient._REGISTRY[register_name] = cls

    @classmethod
    def get_dbclient_class(cls, name):
        return WorkflowDbClient._REGISTRY.get(name, None)

    @classmethod
    def register_actorinfo_filter(cls, method: Callable[[dict], dict]):
        if method not in cls._ACTORINFO_FILTERS:
            WorkflowDbClient._ACTORINFO_FILTERS.append(method)

    @classmethod
    def apply_actorinfo_filters(cls, info: dict) -> dict:
        for method in WorkflowDbClient._ACTORINFO_FILTERS:
            info = method(info)
        return info

    def connect(self, *args, **kw):
        raise NotImplementedError

    def disconnect(self, *args, **kw):
        raise NotImplementedError

    def startWorkflow(self, name: str):
        raise NotImplementedError

    def endWorkflow(self, status="finished") -> None:
        raise NotImplementedError

    def updateWorkflowInfo(self, info: dict) -> None:
        raise NotImplementedError

    def setWorkflowStatus(self, status: str) -> None:
        self.updateWorkflowInfo({"status": status})

    def getWorkflowInfo(self) -> Optional[dict]:
        raise NotImplementedError

    def startActor(
        self, name: str, info: Optional[dict] = None, script: Optional[str] = None
    ):
        raise NotImplementedError

    def endActor(self, actorId, status="finished") -> None:
        raise NotImplementedError

    def updateActorInfo(self, actorId, info: dict) -> None:
        raise NotImplementedError

    def setActorStatus(self, actorId, status: str) -> None:
        self.updateActorInfo(actorId, {"status": status})

    def getActorInfo(self, actorId) -> Optional[None]:
        raise NotImplementedError



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import unittest

from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ErrorHandler import ErrorHandler
from pypushflow.ForkActor import ForkActor
from pypushflow.JoinActor import JoinActor
from pypushflow.RouterActor import RouterActor
from pypushflow.ThreadCounter import ThreadCounter


class TestPythonActor(unittest.TestCase):
    def setUp(self):
        self.thread_counter = ThreadCounter()

    def test_PythonActor(self):
        script = "pypushflow.test.pythonActorTest.py"
        name = "Python Actor Test"
        actor = PythonActor(
            script=script, name=name, thread_counter=self.thread_counter
        )
        stopActor = StopActor(thread_counter=self.thread_counter)
        inData = {"name": "Ragnar"}
        actor.connect(stopActor)
        actor.trigger(inData)
        stopActor.join(timeout=10)
        outData = stopActor.outData
        self.assertIsNotNone(outData)
        self.assertEqual(outData["reply"], "Hello Ragnar!")

    def test_ErrorHandler(self):
        return
        script = "pypushflow.test.pythonErrorHandlerTest.py"
        name = "Python Error Handler Test"
        actor = PythonActor(
            script=script, name=name, thread_counter=self.thread_counter
        )
        stopActor = StopActor(thread_counter=self.thread_counter)
        errorHandler = ErrorHandler(
            name="Error handler", thread_counter=self.thread_counter
        )
        inData = {"name": "Ragnar"}
        actor.connect(stopActor)
        actor.connectOnError(errorHandler)
        errorHandler.connect(stopActor)
        actor.trigger(inData)
        stopActor.join(timeout=5)
        outData = stopActor.outData
        self.assertIsNotNone(outData)
        self.assertTrue("WorkflowException" in outData)

    def test_ForkAndJoinActors(self):
        start = StartActor(thread_counter=self.thread_counter)
        stop = StopActor(thread_counter=self.thread_counter)
        fork = ForkActor(thread_counter=self.thread_counter)
        joinActor = JoinActor(thread_counter=self.thread_counter)
        pythonActor1 = PythonActor(
            script="pypushflow.test.pythonActor1.py", thread_counter=self.thread_counter
        )
        pythonActor2 = PythonActor(
            script="pypushflow.test.pythonActor2.py", thread_counter=self.thread_counter
        )
        # Connections
        start.connect(fork)
        fork.connect(pythonActor1)
        fork.connect(pythonActor2)
        pythonActor1.connect(joinActor)
        joinActor.increaseNumberOfThreads()
        pythonActor2.connect(joinActor)
        joinActor.increaseNumberOfThreads()
        joinActor.connect(stop)
        # Run
        inData = {"a": 1}
        start.trigger(inData)
        stop.join(timeout=5)
        outData = stop.outData
        self.assertIsNotNone(outData)

    def test_RouterActor(self):
        start = StartActor(thread_counter=self.thread_counter)
        stop = StopActor(thread_counter=self.thread_counter)
        router = RouterActor(
            parent=None,
            itemName="a",
            listPort=["other", "null"],
            thread_counter=self.thread_counter,
        )
        pythonActor1 = PythonActor(
            script="pypushflow.test.pythonActor1.py", thread_counter=self.thread_counter
        )
        pythonActor2 = PythonActor(
            script="pypushflow.test.pythonActor2.py", thread_counter=self.thread_counter
        )
        # Connections
        start.connect(router)
        router.connect(pythonActor1, "other")
        router.connect(pythonActor2, "null")
        pythonActor1.connect(stop)
        pythonActor2.connect(stop)
        # Run
        inData = {"a": 1}
        start.trigger(inData)
        stop.join(timeout=5)
        outData = stop.outData
        self.assertIsNotNone(outData)
        self.assertTrue(outData["actor1"])
        # Run 2
        inData = {"a": None}
        start.trigger(inData)
        stop.join(timeout=5)
        outData = stop.outData
        self.assertIsNotNone(outData)
        self.assertTrue(outData["actor2"])



from typing import Optional

try:
    from bson.objectid import ObjectId
    from mongita import MongitaClientMemory
except Exception:
    ObjectId = None
    MongitaClientMemory = None
from .mongo import MongoWorkflowDbClient


class MemoryWorkflowDbClient(MongoWorkflowDbClient, register_name="memory"):
    """Client of an in-memory Mongo database for storing workflow executions.
    Used for testing purposes.
    """

    def connect(self):
        if MongitaClientMemory is None:
            return
        self._collection = MongitaClientMemory()["ppf"]["ppf"]

    def disconnect(self, *args, **kw):
        self._collection = None

    def generateWorkflowId(self, oid: Optional[str] = None) -> ObjectId:
        return ObjectId(oid=oid)

    def generateActorId(self, oid: Optional[str] = None) -> ObjectId:
        return ObjectId(oid=oid)

    def _appendActorInfo(self, actorInfo: dict):
        workflowInfo = self._workflowInfo
        workflowInfo["actors"].append(actorInfo)
        self._workflowInfo = workflowInfo



"""Persistent recording of a workflow executions.

When the "PYPUSHFLOW_MONGOURL" environment variable is specified,
the default database client is the PyBes client. Otherwise the
default client is a dummy client.
"""

import os
from typing import Callable
from .interface import WorkflowDbClient
from .pymongo import PyMongoWorkflowDbClient  # noqa F401
from .pybes import PyBesWorkflowDbClient  # noqa F401
from .mongita import MemoryWorkflowDbClient  # noqa F401
from .dummy import DummyWorkflowDbClient  # noqa F401

DEFAULT_DB_TYPE = "dummy"


def db_client(*args, db_type=None, **kwargs) -> WorkflowDbClient:
    if db_type is None:
        if os.environ.get("PYPUSHFLOW_MONGOURL"):
            db_type = "pybes"
        else:
            db_type = DEFAULT_DB_TYPE
    db_client_class = WorkflowDbClient.get_dbclient_class(db_type)
    client = db_client_class(*args, **kwargs)
    client.connect()
    return client


def register_actorinfo_filter(method: Callable):
    WorkflowDbClient.register_actorinfo_filter(method)



from typing import Optional
from .interface import WorkflowDbClient


class DummyWorkflowDbClient(WorkflowDbClient, register_name="dummy"):
    """Client without a database. Used for testing purposes."""

    def connect(self, *args, **kw):
        pass

    def disconnect(self, *args, **kw):
        pass

    def startWorkflow(self, name: str):
        pass

    def endWorkflow(self, status="finished") -> None:
        pass

    def updateWorkflowInfo(self, info: dict) -> None:
        pass

    def getWorkflowInfo(self) -> Optional[dict]:
        pass

    def startActor(
        self, name: str, info: Optional[dict] = None, script: Optional[str] = None
    ):
        pass

    def endActor(self, actorId, status="finished") -> None:
        pass

    def updateActorInfo(self, actorId, info: dict) -> None:
        pass

    def getActorInfo(self, actorId) -> Optional[None]:
        pass



import unittest
import os


class PyBesTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ[
            "PYPUSHFLOW_MONGOURL"
        ] = "mongodb://pybes:pybes@linsvensson.esrf.fr:27017/pybes"
        os.environ["PYPUSHFLOW_CREATOR"] = "PyPushflowUnitTests"

    def tearDown(self) -> None:
        os.environ.pop("PYPUSHFLOW_MONGOURL", None)
        os.environ.pop("PYPUSHFLOW_CREATOR", None)



def run(value=None, limit=10, **kwargs):
    if value is None:
        raise RuntimeError("Missing argument 'value'!")
    if limit is None:
        raise RuntimeError("Missing argument 'limit'!")
    doContinue = "true"
    if value >= limit:
        doContinue = "false"
    return {"doContinue": doContinue}



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"


import pprint
from pypushflow.ThreadCountingActor import ThreadCountingActor
from pypushflow.logutils import PyPushflowLoggedObject


class Port(ThreadCountingActor):
    def __init__(self, errorHandler, name, **kw):
        super().__init__(name=errorHandler.name + "." + name, **kw)
        self.errorHandler = errorHandler
        self.listActor = []
        self.inPortTrigger = None

    def connect(self, actor):
        self.logger.debug("connectig to '%s'", actor.name)
        self.listActor.append(actor)

    def setTrigger(self, trigger):
        self.inPortTrigger = trigger

    def trigger(self, inData):
        self.logger.info(
            "triggered with inData =\n %s",
            pprint.pformat(inData),
        )
        if len(self.listActor) > 0:
            for actor in self.listActor:
                self.logger.debug(
                    "In trigger '%s' -> actorName '%s'",
                    self.errorHandler.name,
                    actor.name,
                )
                actor.trigger(inData)
        if self.inPortTrigger is not None:
            self.logger.debug(
                "In '%s' trigger, trigger = '%s'",
                self.errorHandler.name,
                self.inPortTrigger,
            )
            self.inPortTrigger(inData)


class Submodel(PyPushflowLoggedObject):
    def __init__(
        self,
        parent=None,
        errorHandler=None,
        name=None,
        portNames=["In", "Out"],
        thread_counter=None,
    ):
        super().__init__(log_metadata={"submodel": name}, parent=parent)
        self.parent = parent
        self.name = name
        if errorHandler is None:
            self.errorHandler = parent
        else:
            self.errorHandler = errorHandler
        self.dictPort = {}
        self.listOnErrorActor = []
        for portName in portNames:
            self.dictPort[portName] = Port(
                self, portName, thread_counter=thread_counter
            )

    @property
    def db_client(self):
        return self.parent.db_client

    def setStatus(self, status):
        self.parent.setStatus(status)

    def getActorPath(self):
        return self.parent.getActorPath() + "/" + self.name.replace("%", " ")

    def getPort(self, portName):
        return self.dictPort[portName]

    def connect(self, actor, portName="Out"):
        self.dictPort[portName].connect(actor)

    def connectOnError(self, actor):
        self.logger.debug("connect to error handler '%s'", actor.name)
        self.listOnErrorActor.append(actor)

    def triggerOnError(self, inData):
        self.logger.info(
            "triggered due to error with inData =\n %s", pprint.pformat(inData)
        )
        for onErrorActor in self.listOnErrorActor:
            onErrorActor.trigger(inData)
        if self.errorHandler is not None:
            self.errorHandler.triggerOnError(inData)

    def addActorRef(self, actorRef):
        if self.parent is not None:
            self.parent.addActorRef(actorRef)

    @property
    def pool(self):
        if self.parent is not None:
            return self.parent.pool



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "07/04/2021"

from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.test.workflowTestCase import WorkflowTestCase


class Workflow1(Workflow):
    """
    Workflow containing one start actor,
    one python actor and one stop actor.
    """

    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(parent=self, thread_counter=ctr)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.test.pythonActorTest.py",
            name="Python Actor Test",
            thread_counter=ctr,
        )
        self.stopActor = StopActor(parent=self, thread_counter=ctr)
        self.startActor.connect(self.pythonActor)
        self.pythonActor.connect(self.stopActor)


class TestWorkflow1(WorkflowTestCase):
    def test_Workflow1(self):
        testWorkflow1 = Workflow1("Test workflow 1")
        inData = {"name": "Jerry"}
        outData = testWorkflow1.run(inData, timeout=15, shared_pool=True)
        self.assertIsNotNone(outData)
        self.assertEqual(outData["reply"], "Hello Jerry!")



from typing import Optional

try:
    from bson.objectid import ObjectId
    from pymongo import MongoClient
except Exception:
    ObjectId = None
    MongoClient = None
from .mongo import MongoWorkflowDbClient


class PyMongoWorkflowDbClient(MongoWorkflowDbClient, register_name="pymongo"):
    """Client of an external Mongo database for storing workflow executions."""

    def connect(self, url: str, database: str, collection: str):
        if MongoClient is None:
            return
        client = MongoClient(url, serverSelectionTimeoutMS=1000)
        self._client = client
        self._collection = client[database][collection]

    def disconnect(self, *args, **kw):
        self._collection = None
        if self._client is not None:
            self._client.close()
            self._client = None

    def generateWorkflowId(self, oid: Optional[str] = None) -> ObjectId:
        return ObjectId(oid=oid)

    def generateActorId(self, oid: Optional[str] = None) -> ObjectId:
        return ObjectId(oid=oid)

    def _appendActorInfo(self, actorInfo: dict):
        self._collection.update_one(
            {"_id": self._workflowId}, {"$push": {"actors": actorInfo}}
        )



import psutil
import unittest
import logging

logger = logging.getLogger(__name__)


class WorkflowTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self._current_process = psutil.Process()
        self._subprocesses = {
            child.pid for child in self._current_process.children(recursive=True)
        }
        self._fail_on_subprocesses = True

    def tearDown(self) -> None:
        has_new_subprocesses = False
        for child in self._current_process.children(recursive=True):
            if child.pid not in self._subprocesses:
                has_new_subprocesses = True
                logger.error("dangling child %s", child)
                try:
                    child.terminate()
                except psutil.NoSuchProcess:
                    pass
        if has_new_subprocesses and self._fail_on_subprocesses:
            raise RuntimeError("not all subprocesses have finished")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "03/05/2021"

from pypushflow.Workflow import Workflow
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
from pypushflow.RouterActor import RouterActor
from pypushflow.ThreadCounter import ThreadCounter
from pypushflow.test.workflowTestCase import WorkflowTestCase


class Workflow10(Workflow):
    def __init__(self, name):
        super().__init__(name)
        ctr = ThreadCounter(parent=self)
        self.startActor = StartActor(self, thread_counter=ctr)
        self.pythonActorAddWithoutSleep = PythonActor(
            parent=self,
            script="pypushflow.test.pythonActorAddWithoutSleep.py",
            name="Add without sleep",
            thread_counter=ctr,
        )
        self.pythonActorCheck = PythonActor(
            parent=self,
            script="pypushflow.test.pythonActorCheck.py",
            name="Check",
            thread_counter=ctr,
        )
        self.check = RouterActor(
            parent=self,
            name="Check",
            itemName="doContinue",
            listPort=["true", "false"],
            thread_counter=ctr,
        )
        self.stopActor = StopActor(self, thread_counter=ctr)
        self.startActor.connect(self.pythonActorAddWithoutSleep)
        self.pythonActorAddWithoutSleep.connect(self.pythonActorCheck)
        self.pythonActorCheck.connect(self.check)
        self.check.connect(self.pythonActorAddWithoutSleep, expectedValue="true")
        self.check.connect(self.stopActor, expectedValue="false")


class TestWorkflow10(WorkflowTestCase):
    def test_workflow10(self):
        limit = 10
        workflow10 = Workflow10(f"Test workflow {limit}")
        inData = {"value": 1, "limit": limit}
        outData = workflow10.run(inData, timeout=1200, shared_pool=True)
        self.assertEqual(outData["value"], limit)



import sys
import unittest
import itertools
from threading import Event
from contextlib import contextmanager
from pypushflow.AsyncExecute import ProcessPool
from pypushflow.AsyncExecute import apply_async


def add(a, b):
    return a + b


def error(a, b):
    raise RuntimeError("intentional error")


class TestAsyncExecute(unittest.TestCase):
    def setUp(self) -> None:
        self.callback_event = Event()
        self.failed_msg = ""

    def reset(self):
        self.callback_event.clear()
        self.failed_msg = ""

    @property
    def contexts(self):
        if sys.platform != "win32":
            return [None, "fork", "spawn", "forkserver"]
        else:
            return [None, "spawn"]

    def test_async_execution(self):
        context = self.contexts
        daemon = [True, False]
        manage_pool = [True, False]
        functions = [add, error]
        for context, daemon, manage_pool, func in itertools.product(
            context, daemon, manage_pool, functions
        ):
            pool_options = {
                "context": context,
                "daemon": daemon,
                "manage_pool": manage_pool,
            }
            with self.subTest(func=func.__name__, **pool_options):
                self.run_test(func, **pool_options)

    @contextmanager
    def run_context(self, manage_pool=True, **pool_options):
        self.reset()
        if manage_pool:
            pool = None
            try:
                with ProcessPool(processes=1, **pool_options) as pool:
                    yield {"pool": pool}
            finally:
                if pool is not None:
                    pool.join()
        else:
            yield pool_options

    def run_test(self, func, **pool_options):
        with self.run_context(**pool_options) as pool_options:
            apply_async(
                func,
                args=(1, 1),
                callback=self._result_callback,
                error_callback=self._error_callback,
                **pool_options,
            )
            self.callback_event.wait(5)
            self.assertFalse(bool(self.failed_msg), msg=self.failed_msg)

    def _result_callback(self, return_value):
        try:
            if return_value != 2:
                self.failed_msg = f"{return_value} != 2"
        finally:
            self.callback_event.set()

    def _error_callback(self, exception):
        try:
            if not isinstance(exception, RuntimeError):
                self.failed_msg = f"{exception} is not a RuntimeError"
            elif str(exception) != "intentional error":
                self.failed_msg = f"'{exception}' != 'intentional error'"
        finally:
            self.callback_event.set()



import os
import sys
import logging
import logging.handlers
from pathlib import Path

DEFAULT_FORMAT = "%(levelname)-8s %(asctime)s [%(threadName)s] %(message)s"
DEFAULT_LOGDIR = os.path.join(os.path.sep, "tmp_14_days")


class PyPushflowLoggedObject:
    """This object owns a logger which adds workflow related context to the logs"""

    def __init__(self, log_metadata=None, parent=None):
        if log_metadata:
            if "class" not in log_metadata:
                log_metadata["class"] = type(self).__name__
        else:
            log_metadata = {"class": type(self).__name__}
        if parent:
            logger = parent.logger
        else:
            logger = logging.getLogger("pypushflow")
        self.logger = _PypushflowLoggerAdapter(logger, log_metadata)


def basicConfig(filename=None, logdir=None, **kwargs):
    """Basic configuration for the root logger. By default:
        * set log level to DEBUG
        * add STDOUT stream handler

    Has no effect when there are already handlers unless `force=True`.
    Note: pytest installs handlers.
    """
    kwargs.setdefault("format", DEFAULT_FORMAT)
    kwargs.setdefault("level", logging.DEBUG)
    kwargs.setdefault("stream", sys.stdout)
    force = kwargs.setdefault("force", False)
    logging.basicConfig(**kwargs)
    logger = logging.getLogger()
    for handler in logger.handlers:
        if isinstance(handler, _PyPushflowLogFile):
            logger.removeHandler(handler)
    if (len(logger.handlers) == 0 or force) and filename:
        if logdir is None:
            logdir = DEFAULT_LOGDIR
        filename = os.path.join(logdir, filename)
        handler = _PyPushflowLogFile.factory(filename)
        logger.addHandler(handler)


class _PypushflowLoggerAdapter(logging.LoggerAdapter):
    """Log adapter which adds workflow related context (if any) to the log messages"""

    def process(self, msg, kwargs):
        kwargs["extra"] = self.extra

        if self.extra:
            if "workflow" in self.extra:
                ctx = "[%s]" % (self.extra["workflow"],)
            elif "actor" in self.extra:
                ctx = "[<%s> %s]" % (
                    self.extra["class"],
                    self.extra["actor"],
                )
            else:
                ctx = None
            if ctx:
                msg = "%s %s" % (ctx, msg)
        return msg, kwargs


class _PyPushflowLogFile(logging.handlers.RotatingFileHandler):
    MAXBYTES = 1e7
    BACKUPCOUNT = 10
    SUFFIX = ".log"
    DEFAULT_USER = "unknown"
    DEFAULT_INITIATOR = "pypushflow"

    @classmethod
    def factory(cls, filename):
        log_file_path = cls.__getLogFile(filename)
        if log_file_path is None:
            return None
        handler = cls(log_file_path, maxBytes=cls.MAXBYTES, backupCount=cls.BACKUPCOUNT)
        formatter = logging.Formatter(DEFAULT_FORMAT)
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)
        return handler

    @classmethod
    def __getLogFile(cls, filename):
        if not filename:
            return None
        user = os.environ.get("USER", cls.DEFAULT_USER)
        initiator = os.environ.get("PYPUSHFLOW_INITIATOR", cls.DEFAULT_INITIATOR)
        filename = Path(filename).with_suffix(cls.SUFFIX)
        directory = filename.parent / user / initiator
        try:
            directory.mkdir(mode=0o755, parents=True, exist_ok=True)
        except Exception:
            return None
        return directory / filename.name



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import pprint
from pypushflow import Submodel
from pypushflow.ThreadCountingActor import ThreadCountingActor


class StopActor(ThreadCountingActor):
    def __init__(self, parent=None, errorHandler=None, name="Stop actor", **kw):
        super().__init__(name=name, parent=parent, **kw)
        self.errorHandler = errorHandler
        self.outData = None

    def trigger(self, inData):
        self.logger.info("triggered with inData =\n %s", pprint.pformat(inData))
        if self.parent is not None and not isinstance(self.parent, Submodel.Submodel):
            # Parent is a Workflow
            self.outData = inData
        elif self.errorHandler is not None:
            self.errorHandler.errorHandler.stopActor.trigger(inData)
        else:
            self.outData = inData

    def join(self, timeout=7200):
        if self.parent is not None:
            self.logger.debug("wait for scheduler threads to be finished")
        success = self._wait_threads_finished(timeout=timeout)
        if self.parent is not None:
            self.logger.debug("scheduler threads are finished")
        self._finalizeInMongo(success)
        return success

    def _finalizeInMongo(self, success):
        if self.parent is None:
            return
        if success:
            self.logger.debug("finished")
            self.parent.setStatus("finished")
        else:
            self.logger.error("timeout detected")
            self.parent.setStatus("timeout")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import os
import time
import pprint
import datetime
import traceback
import importlib
import multiprocessing
import multiprocessing.pool

from pypushflow.AbstractActor import AbstractActor
from pypushflow.AsyncExecute import apply_async


class PythonActor(AbstractActor):
    def __init__(
        self,
        parent=None,
        name="Python Actor",
        errorHandler=None,
        script=None,
        daemon=False,
        context=None,
        **kw
    ):
        super().__init__(parent=parent, name=name, **kw)
        self.parentErrorHandler = errorHandler
        self.listErrorHandler = []
        self.script = script
        self.inData = None
        self.pool_options = {"daemon": daemon, "context": context}

    def connectOnError(self, actor):
        self.logger.debug("connect to error handler '%s'", actor.name)
        self.listErrorHandler.append(actor)

    def trigger(self, inData: dict):
        self.logger.info("triggered with inData =\n %s", pprint.pformat(inData))
        self.setStarted()
        self.inData = dict(inData)
        self.uploadInDataToMongo(actorData={"inData": inData}, script=self.script)

        try:
            module = importlib.import_module(os.path.splitext(self.script)[0])
        except Exception as e:
            self.logger.error("Error when trying to import script '%s'", self.script)
            time.sleep(1)
            self.errorHandler(e)
            return

        with self._postpone_end_thread(self.resultHandler, self.errorHandler) as (
            resultHandler,
            errorHandler,
        ):
            target = module.run
            self.logger.debug(
                "asynchronous execution of '%s.%s'",
                target.__module__,
                target.__name__,
            )
            apply_async(
                target,
                kwds=self.inData,
                callback=resultHandler,
                error_callback=errorHandler,
                pool=self.pool,
                **self.pool_options,
            )

    @property
    def pool(self):
        if self.parent is not None:
            return self.parent.pool

    @property
    def pool_resources(self):
        return 1

    def resultHandler(self, result: dict):
        """Async callback in case of success"""
        try:
            # Handle the result
            self._finishedSuccess(result)

            # Trigger actors
            downstreamData = dict(self.inData)
            downstreamData.update(result)
            self._triggerDownStreamActors(downstreamData)
        except Exception as e:
            self.errorHandler(e)

    def errorHandler(self, exception: Exception):
        """Async callback in case of exception"""
        try:
            # Handle the result
            self._logException(exception)
            result = self._parseException(exception)
            self._finishedFailure(result)

            # Trigger actors
            downstreamData = dict(self.inData)
            downstreamData["WorkflowException"] = result
            self._triggerErrorHandlers(downstreamData)
        except Exception:
            self.logger.exception("In errorHandler for '%s'", self.name)

    def _logException(self, exception: Exception) -> None:
        if isinstance(exception.__cause__, multiprocessing.pool.RemoteTraceback):
            exception = exception.__cause__
            logfunc = self.logger.error
        elif isinstance(exception, multiprocessing.pool.MaybeEncodingError):
            # This exception has no traceback
            logfunc = self.logger.error
        else:
            logfunc = self.logger.exception
        logfunc(
            "Error in python actor '%s'!\n Not running down stream actors %s\n Exception:%s",
            self.name,
            [actor.name for actor in self.listDownStreamActor],
            exception,
        )

    def _parseException(self, exception: Exception) -> dict:
        errorMessage = str(exception)
        if isinstance(exception.__cause__, multiprocessing.pool.RemoteTraceback):
            exception = exception.__cause__
        traceBack = traceback.format_exception(
            type(exception), exception, exception.__traceback__
        )
        return {
            "errorMessage": errorMessage,
            "traceBack": traceBack,
        }

    def _triggerDownStreamActors(self, downstreamData: dict):
        for downStreamActor in self.listDownStreamActor:
            self.logger.debug(
                "trigger actor '%s' with inData =\n %s",
                downStreamActor.name,
                pprint.pformat(downstreamData),
            )
            downStreamActor.trigger(downstreamData)

    def _triggerErrorHandlers(self, downstreamData: dict):
        for errorHandler in self.listErrorHandler:
            errorHandler.trigger(downstreamData)
        if self.parentErrorHandler is not None:
            self.parentErrorHandler.triggerOnError(inData=downstreamData)

    def _finishedSuccess(self, result: dict):
        self.setFinished()
        self.uploadOutDataToMongo(
            actorData={
                "stopTime": datetime.datetime.now(),
                "status": "finished",
                "outData": result,
            }
        )
        if "workflowLogFile" in result:
            self.setMongoAttribute("logFile", result["workflowLogFile"])
        if "workflowDebugLogFile" in result:
            self.setMongoAttribute("debugLogFile", result["workflowDebugLogFile"])

    def _finishedFailure(self, result: dict):
        self.setFinished()
        self.uploadOutDataToMongo(
            actorData={
                "stopTime": datetime.datetime.now(),
                "status": "error",
                "outData": result,
            }
        )



import unittest
from pypushflow.persistence import db_client
from .pybesTestCase import PyBesTestCase


class PersistenceTests:
    def setUp(self):
        super().setUp()
        self.skip_message = ""

    def test_Workflow(self):
        name = "test_startWorkflow"
        self.db_client.startWorkflow(name=name)
        info = self.db_client.getWorkflowInfo()
        if info is None:
            self.skipTest(self.skip_message)
            return
        self.assertEqual(info["status"], "started")

        self.db_client.updateWorkflowInfo({"status": "error"})
        status = self.db_client.getWorkflowInfo()["status"]
        self.assertEqual(status, "error")

        self.db_client.endWorkflow()
        status = self.db_client.getWorkflowInfo()["status"]
        self.assertEqual(status, "error")

    def test_Actor(self):
        name = "test_startWorkflow"
        self.db_client.startWorkflow(name=name)
        actorName1 = "TestActor1"

        actorId1 = self.db_client.startActor(actorName1)
        info = self.db_client.getActorInfo(actorId1)
        if info is None:
            self.skipTest(self.skip_message)
            return
        self.assertEqual(info["status"], "started")

        self.db_client.updateActorInfo(actorId1, {"status": "error"})
        status = self.db_client.getActorInfo(actorId1)["status"]
        self.assertEqual(status, "error")

        self.db_client.endActor(actorId1)
        status = self.db_client.getActorInfo(actorId1)["status"]
        self.assertEqual(status, "error")

        actorName2 = "TestActor2"
        actorId2 = self.db_client.startActor(name=actorName2)
        status = self.db_client.getActorInfo(actorId2)["status"]
        self.assertEqual(status, "started")

        self.db_client.updateActorInfo(actorId2, {"data": {"a": 1}})
        data = self.db_client.getActorInfo(actorId2)["data"]
        self.assertEqual(data, {"a": 1})

        self.db_client.endActor(actorId2)
        status = self.db_client.getActorInfo(actorId2)["status"]
        self.assertEqual(status, "finished")


class TestDummyPersistence(unittest.TestCase, PersistenceTests):
    def setUp(self):
        super().setUp()
        self.skip_message = "dummy"
        self.db_client = db_client(db_type="dummy")
        self.db_client.connect()


class TestMemoryPersistence(unittest.TestCase, PersistenceTests):
    def setUp(self):
        super().setUp()
        self.skip_message = "requires the 'mongita' library"
        self.db_client = db_client(db_type="memory")
        self.db_client.connect()


class TestPyBesPersistence(PyBesTestCase, PersistenceTests):
    def setUp(self):
        super().setUp()
        self.skip_message = "requires external database and the 'pymongo' library"
        self.db_client = db_client(db_type="pybes")
        self.db_client.connect()



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"


import pprint
import logging
from contextlib import contextmanager
from pypushflow.persistence import db_client
from pypushflow.logutils import PyPushflowLoggedObject
from pypushflow.logutils import basicConfig
from pypushflow.AsyncExecute import ProcessPool


class Workflow(PyPushflowLoggedObject):
    def __init__(self, name, daemon=False, context=None, level=logging.DEBUG):
        basicConfig(filename=name, level=level)
        super().__init__(log_metadata={"workflow": name})
        self.logger.info("\n\nStarting new workflow '%s'\n", name)
        self.name = name
        self.listOnErrorActor = []
        self.db_client = db_client()
        self.db_client.startWorkflow(name)
        self.listActorRef = []
        self.pool_options = {"daemon": daemon, "context": context}
        self.__pool = None

    def connectOnError(self, actor):
        self.logger.debug("connect to error handler '%s'", actor.name)
        self.listOnErrorActor.append(actor)

    def triggerOnError(self, inData):
        self.logger.info(
            "triggered due to error with inData =\n %s", pprint.pformat(inData)
        )
        for onErrorActor in self.listOnErrorActor:
            onErrorActor.trigger(inData)

    def getActorPath(self):
        return "/" + self.name

    def addActorRef(self, actorRef):
        self.logger.debug("add reference to actor '%s'", actorRef.name)
        self.listActorRef.append(actorRef)

    def getListActorRef(self):
        return self.listActorRef

    def setStatus(self, status):
        self.db_client.setWorkflowStatus(status)

    @property
    def pool(self):
        return self.__pool

    @contextmanager
    def _run_context(self, shared_pool=False):
        if not shared_pool:
            # A new pool with 1 worker will be created for each asynchronous
            # call and the pool cleanup is done in the callbacks
            yield
            return
        if self.__pool is not None:
            # A pool already exists
            yield
            return
        processes = sum(actor.pool_resources for actor in self.listActorRef)
        try:
            with ProcessPool(processes=processes, **self.pool_options) as pool:
                self.__pool = pool
                yield
        finally:
            if self.__pool is not None:
                self.__pool.join()
                self.__pool = None

    def run(self, inData, timeout=None, shared_pool=False):
        with self._run_context(shared_pool=shared_pool):
            self.startActor.trigger(inData)
            self.stopActor.join(timeout=timeout)
            return self.stopActor.outData



import unittest
from time import sleep
from pypushflow.ThreadCounter import ThreadCounter
from concurrent.futures import ThreadPoolExecutor


class TestThreadCounter(unittest.TestCase):
    def setUp(self):
        self.counter = ThreadCounter()

    def mythread(self, sleep_time):
        with self.counter:
            sleep(1)
        return True

    def test_thread_count(self):
        self.assertEqual(self.counter.nthreads, 0)
        with self.counter as ctr:
            self.assertEqual(ctr.nthreads, 1)
            with self.counter as ctr:
                self.assertEqual(ctr.nthreads, 2)
            self.assertEqual(ctr.nthreads, 1)
        self.assertEqual(self.counter.nthreads, 0)

    def test_multiple_threads(self):
        with ThreadPoolExecutor(max_workers=10) as executor:
            results = executor.map(self.mythread, [1] * 10)
            self.counter.wait_threads_finished()
            for result in results:
                self.assertTrue(result)



from functools import wraps
from contextlib import contextmanager
from pypushflow.logutils import PyPushflowLoggedObject
from pypushflow.ActorInterface import ActorInterface


def with_thread_context(trigger):
    """Wraps the `trigger` method of all derived classes of ThreadCountingActor"""

    @wraps(trigger)
    def wrapper(self, *args, **kw):
        with self._thread_context():
            return trigger(self, *args, **kw)

    return wrapper


def callback_with_end_thread(async_callback, end_thread, log_msg):
    """Wraps a async_callback"""

    @wraps(async_callback)
    def wrapper(*args, **kw):
        try:
            return async_callback(*args, **kw)
        finally:
            end_thread(msg=log_msg)

    return wrapper


class ThreadCountingActor(PyPushflowLoggedObject, ActorInterface):
    """The `trigger` method will increase the thread counter
    at the start and decrease the thread counter at the end.
    """

    def __init__(self, name=None, parent=None, thread_counter=None):
        if name is None:
            raise RuntimeError("Actor name is None!")
        if thread_counter is None:
            raise ValueError("Actor requires a 'thread_counter' argument")
        super().__init__(log_metadata={"actor": name}, parent=parent)
        self.name = name
        self.parent = parent
        if parent is not None:
            parent.addActorRef(self)
        self.__thread_counter = thread_counter
        self.__in_thread_context = False
        self.__postpone_end_thread = False

    def __init_subclass__(subcls, **kw):
        """Wrap the `trigger` method"""
        super().__init_subclass__(**kw)
        subcls.trigger = with_thread_context(subcls.trigger)

    @contextmanager
    def _thread_context(self):
        """Re-entrant context manager that starts a thread
        on first entrance and ends a thread on last exit,
        unless the thread ending is post-poned until after
        and async callback.
        """
        if self.__in_thread_context:
            yield
            return
        self.__thread_counter.start_thread(msg="Thread started for " + repr(self.name))
        try:
            self.__in_thread_context = True
            self.__postpone_end_thread = False
            try:
                yield
            finally:
                self.__in_thread_context = False
        finally:
            if self.__postpone_end_thread:
                self.__postpone_end_thread = False
            else:
                self.__thread_counter.end_thread(
                    msg="Thread ended for " + repr(self.name)
                )

    @contextmanager
    def _postpone_end_thread(self, *async_callbacks):
        """Post-pone thread ending until after an async callback is executed.

        Only one of the async callbacks is expected to be called.
        """
        if self.__in_thread_context:
            self.__postpone_end_thread = True
        try:
            async_callbacks = tuple(
                callback_with_end_thread(
                    async_callback,
                    self.__thread_counter.end_thread,
                    "Thread ended for " + repr(self.name),
                )
                for async_callback in async_callbacks
            )
            yield async_callbacks
        except BaseException:
            if self.__in_thread_context:
                self.__postpone_end_thread = False
            raise

    def _wait_threads_finished(self, **kw):
        return self.__thread_counter.wait_threads_finished(**kw)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import pprint
from pypushflow.ThreadCountingActor import ThreadCountingActor


class AbstractActor(ThreadCountingActor):
    def __init__(self, parent=None, name=None, **kw):
        super().__init__(name=name, parent=parent, **kw)
        self.listDownStreamActor = []
        self.actorId = None
        self.started = False
        self.finished = False

    def __str__(self) -> str:
        return self.name

    def connect(self, actor):
        self.logger.debug("connect to actor '%s'", actor.name)
        self.listDownStreamActor.append(actor)

    def trigger(self, inData):
        self.logger.info("triggered with inData =\n %s", pprint.pformat(inData))
        self.setStarted()
        self.setFinished()
        for actor in self.listDownStreamActor:
            actor.trigger(inData)

    def uploadInDataToMongo(self, actorData=None, script=None):
        if self.parent is not None:
            name = self.getActorPath() + "/" + self.name
            if actorData:
                info = dict(actorData)
            else:
                info = dict()
            if script:
                info["script"] = script
            self.actorId = self.parent.db_client.startActor(name=name, info=info)

    def uploadOutDataToMongo(self, actorData=None):
        if actorData and self.actorId is not None:
            self.parent.db_client.updateActorInfo(self.actorId, info=actorData)

    def setMongoAttribute(self, attribute, value):
        if self.actorId is not None:
            self.parent.db_client.updateActorInfo(self.actorId, info={attribute: value})

    def getActorPath(self):
        return self.parent.getActorPath()

    def hasStarted(self):
        return self.started

    def setStarted(self):
        self.logger.info("started")
        self.started = True

    def hasFinished(self):
        return self.finished

    def setFinished(self):
        self.logger.info("finished")
        self.finished = True



class ActorInterface:
    """The interface that makes a class a pypushflow actor"""

    def trigger(self, inData):
        raise NotImplementedError

    @property
    def pool_resources(self):
        return 0



import logging
from datetime import datetime
from typing import Optional
from .interface import WorkflowDbClient

logger = logging.getLogger(__name__)


class MongoWorkflowDbClient(WorkflowDbClient):
    """Client interface of a Mongo database for storing workflow executions."""

    def __init__(self):
        self._collection = None
        self._workflowId = None

    def generateWorkflowId(self, *args, **kw):
        raise NotImplementedError

    def generateActorId(self, *args, **kw):
        raise NotImplementedError

    def startWorkflow(self, name: str):
        if self._collection is None:
            return
        if self._workflowId is not None:
            raise RuntimeError("Workflow start already logged")
        workflowInfo = self._generateInitialWorkflowInfo()
        workflowInfo["name"] = name
        workflowInfo["status"] = "started"
        workflowInfo["startTime"] = datetime.now()
        try:
            self._collection.insert_one(workflowInfo)
        except Exception:
            self._collection = None
            logger.exception("Mongo database error")
        self._workflowId = workflowInfo["_id"]

    def endWorkflow(self, status="finished") -> None:
        if self._skip:
            return
        workflowInfo = self._workflowInfo
        if workflowInfo["status"] != "error":
            workflowInfo["status"] = status
        workflowInfo["stopTime"] = datetime.now()
        self._workflowInfo = workflowInfo

    def updateWorkflowInfo(self, info: dict) -> None:
        if self._skip:
            return
        workflowInfo = self._workflowInfo
        workflowInfo.update(info)
        self._workflowInfo = workflowInfo

    def getWorkflowInfo(self) -> Optional[dict]:
        if self._collection is None:
            return
        return self._workflowInfo

    def startActor(self, name: str, info: Optional[str] = None):
        if self._skip:
            return
        actorInfo = self._generateInitialActorInfo()
        actorInfo["name"] = name
        actorInfo["status"] = "started"
        if info:
            actorInfo.update(info)
        actorInfo["startTime"] = datetime.now()
        actorInfo = self.apply_actorinfo_filters(actorInfo)
        self._appendActorInfo(actorInfo)
        return actorInfo["_id"]

    def _appendActorInfo(self, actorInfo: dict):
        raise NotImplementedError

    def endActor(self, actorId, status="finished") -> None:
        if self._skip:
            return
        workflowInfo = self._workflowInfo
        for actorInfo in workflowInfo["actors"]:
            if actorInfo["_id"] == actorId:
                if actorInfo["status"] != "error":
                    actorInfo["status"] = status
                actorInfo["stopTime"] = datetime.now()
                self._workflowInfo = workflowInfo
                break

    def updateActorInfo(self, actorId, info: dict) -> None:
        if self._skip:
            return
        info = self.apply_actorinfo_filters(info)
        workflowInfo = self._workflowInfo
        for actorInfo in workflowInfo["actors"]:
            if actorInfo["_id"] == actorId:
                actorInfo.update(info)
                self._workflowInfo = workflowInfo
                break

    def getActorInfo(self, actorId) -> Optional[None]:
        if self._skip:
            return
        workflowInfo = self._workflowInfo
        for actorInfo in workflowInfo["actors"]:
            if actorInfo["_id"] == actorId:
                return actorInfo

    @property
    def _skip(self):
        if self._collection is None:
            return True
        if self._workflowId is None:
            raise RuntimeError("Workflow start not logged")
        return False

    def _generateInitialWorkflowInfo(self) -> dict:
        oid = self.generateWorkflowId()
        return {
            "_id": oid,
            "Request ID": str(oid),
            "name": "unknown",
            "status": "unknown",
            "actors": [],
        }

    def _generateInitialActorInfo(self) -> dict:
        oid = self.generateActorId()
        return {
            "_id": oid,
            "name": "unknown",
            "status": "unknown",
        }

    @property
    def _workflowInfo(self) -> dict:
        return self._collection.find_one({"_id": self._workflowId})

    @_workflowInfo.setter
    def _workflowInfo(self, info: dict):
        self._collection.update_one(
            {"_id": self._workflowId}, {"$set": info}, upsert=False
        )



from threading import Condition
from pypushflow.logutils import PyPushflowLoggedObject


class ThreadCounter(PyPushflowLoggedObject):
    """Scheduling thread counter"""

    def __init__(self, parent=None):
        self.__counter = 0
        self.__condition = Condition()
        super().__init__(parent=parent)

    def start_thread(self, msg=None):
        with self.__condition:
            self.__counter += 1
            self._log_counter_change(msg=msg)
            self.__condition.notify_all()

    def end_thread(self, msg=None):
        with self.__condition:
            self.__counter = max(self.__counter - 1, 0)
            self._log_counter_change(msg=msg)
            self.__condition.notify_all()

    def __enter__(self):
        self.start_thread()
        return self

    def __exit__(self, *args):
        self.end_thread()

    def wait_threads_finished(self, timeout=None):
        """Returns False when timeout expires"""
        while True:
            with self.__condition:
                if self.__counter == 0:
                    break
                if not self.__condition.wait(timeout=timeout):
                    return False
        return True

    @property
    def nthreads(self):
        return self.__counter

    def _log_counter_change(self, msg=None):
        if msg is None:
            msg = "Thread counter changed"
        self.logger.debug("%s (%d threads running)", msg, self.__counter)
